package com.maroontress.gcovparser;

import com.maroontress.gcovparser.gcda.FunctionDataRecord;
import com.maroontress.gcovparser.gcno.AnnounceFunctionRecord;
import com.maroontress.gcovparser.gcno.ArcRecord;
import com.maroontress.gcovparser.gcno.ArcsRecord;
import com.maroontress.gcovparser.gcno.FunctionGraphRecord;
import com.maroontress.gcovparser.gcno.LineRecord;
import com.maroontress.gcovparser.gcno.LinesRecord;
import java.util.ArrayList;

/**
   ؿդΥ֥ȥ饯ȼǤ

   @param <T> ֥åζݥ饹
   @param <U> ζݥ饹
*/
public abstract class AbstractFunctionGraph<T extends AbstractBlock<T, U>,
					    U extends AbstractArc<T, U>> {

    /** ̻ҤǤ */
    private int id;

    /** ؿΥåǤ */
    private int checksum;

    /** ؿ̾Ǥ */
    private String functionName;

    /** ɤΥե̾Ǥ */
    private String sourceFile;

    /** ؿи륽ɤιԿǤ */
    private int lineNumber;

    /** ؿܥ֥åΥꥹȤǤ */
    private ArrayList<T> blocks;

    /** ٤ƤΥθĿǤ */
    private int totalArcCount;

    /**
       ΥθĿǤΥϡ㳰longjmp()ˤäơ
       ߤδؿȴϩбޤ
    */
    private int fakeArcCount;

    /** ¹ԲȽƤ륢ΥꥹȤǤ */
    private ArrayList<U> solvedArcs;

    /** ¹ԲʥΥꥹȤǤ */
    private ArrayList<U> unsolvedArcs;

    /** ƤӽФ줿Ǥ */
    private long calledCount;

    /** äǤ */
    private long returnedCount;

    /** ¹Ԥ줿֥åȽиˤθĿǤ */
    private int executedBlockCount;

    /** եդ褷Ƥ뤫ɤɽޤ */
    private boolean solved;

    /**
       ֥åޤ

       @param id ̻
       @param flags ե饰
       @return ֥å
    */
    protected abstract T createBlock(int id, int flags);

    /**
       ޤ

       @param start ϥ֥å
       @param end λ֥å
       @param flags ե饰
       @return 
    */
    protected abstract U createArc(T start, T end, int flags);

    /**
       ֥åIterableޤ

       @return ֥åIterable
    */
    public final Iterable<T> getBlocks() {
	return blocks;
    }

    /**
       Iterableޤ

       @return Iterable
    */
    public final Iterable<U> getArcs() {
	ArrayList<U> arcs = new ArrayList<U>();
	arcs.addAll(solvedArcs);
	arcs.addAll(unsolvedArcs);
	return arcs;
    }

    /**
       åޤ

       @return å
    */
    public final int getChecksum() {
	return checksum;
    }

    /**
       ɤΥե̾ޤ

       @return ɤΥե̾
    */
    public final String getSourceFile() {
	return sourceFile;
    }

    /**
       ʣ٤ޤ

       @return ʣ
    */
    public final int getComplexity() {
	return getComplexityWithFake() - fakeArcCount;
    }

    /**
       եΥθʣ٤ޤ

       @return եΥθʣ
    */
    protected final int getComplexityWithFake() {
	return totalArcCount - blocks.size() + 2;
    }

    /**
       եդ򤤤ɤޤ

       @return եդ򤤤true
    */
    protected final boolean isSolved() {
	return solved;
    }

    /**
       ؿե쥳ɤ饤󥹥󥹤ޤ

       @param rec ؿե쥳
       @throws CorruptedFileException եι¤Ƥ뤳Ȥ򸡽
    */
    protected AbstractFunctionGraph(final FunctionGraphRecord rec)
	throws CorruptedFileException {
	AnnounceFunctionRecord announce = rec.getAnnounce();
	id = announce.getId();
	checksum = announce.getChecksum();
	functionName = announce.getFunctionName();
	sourceFile = announce.getSourceFile();
	lineNumber = announce.getLineNumber();
	solvedArcs = new ArrayList<U>();
	unsolvedArcs = new ArrayList<U>();

	int[] blockFlags = rec.getBlocks().getFlags();
	blocks = new ArrayList<T>(blockFlags.length);
	for (int k = 0; k < blockFlags.length; ++k) {
	    blocks.add(createBlock(k, blockFlags[k]));
	}

	ArcsRecord[] arcs = rec.getArcs();
	for (ArcsRecord e : arcs) {
	    addArcsRecord(e);
	}

	LinesRecord[] lines = rec.getLines();
	for (LinesRecord e : lines) {
	    addLinesRecord(e);
	}

	int blockCount = blocks.size();
	if (blockCount < 2) {
	    throw new CorruptedFileException("lacks entry and/or exit blocks");
	}
	T entryBlock = blocks.get(0);
	T exitBlock = blocks.get(blockCount - 1);
	if (entryBlock.getInArcs().size() != 0) {
	    throw new CorruptedFileException("has arcs to entry block");
	}
	if (exitBlock.getOutArcs().size() != 0) {
	    throw new CorruptedFileException("has arcs from exit block");
	}
	for (T e : blocks) {
	    e.presolve();
	}
    }

    /**
       ARCS쥳ɤ饢ơؿդɲäޤ

       @param arcsRecord ARCS쥳
       @throws CorruptedFileException եι¤Ƥ뤳Ȥ򸡽
    */
    private void addArcsRecord(final ArcsRecord arcsRecord)
	throws CorruptedFileException {
	int blockCount = blocks.size();
	int startIndex = arcsRecord.getStartIndex();
	ArcRecord[] list = arcsRecord.getList();
	if (startIndex >= blockCount) {
	    throw new CorruptedFileException();
	}
	for (ArcRecord arcRecord : list) {
	    int endIndex = arcRecord.getEndIndex();
	    int flags = arcRecord.getFlags();
	    if (endIndex >= blockCount) {
		throw new CorruptedFileException();
	    }
	    T start = blocks.get(startIndex);
	    T end = blocks.get(endIndex);
	    U arc = createArc(start, end, flags);
	    start.addOutArc(arc);
	    end.addInArc(arc);
	    if (!arc.isOnTree()) {
		/*
		  ѥ˥󥰥ĥ꡼ǤϤʤgcdaեˤϤ
		  б¹ԲϿ롣μ¹Բ
		  ºݤ˲褹ΤsetFunctionDataRecord()᥽
		  ɤƤФ줿Ȥ
		*/
		solvedArcs.add(arc);
	    } else {
		/*
		  ѥ˥󥰥ĥ꡼Υμ¹Բgcdaե
		  뤫ǤʤΤǡեդ򤯤Ȥǥ
		  μ¹ԲʤФʤʤ
		*/
		unsolvedArcs.add(arc);
	    }
	    if (arc.isFake()) {
		++fakeArcCount;
	    }
	}
	totalArcCount += list.length;
    }

    /**
       LINES쥳ɤԥȥΥꥹȤơؿդɲ
       ޤ

       @param linesRecord LINES쥳
       @throws CorruptedFileException եι¤Ƥ뤳Ȥ򸡽
    */
    private void addLinesRecord(final LinesRecord linesRecord)
	throws CorruptedFileException {
	int blockIndex = linesRecord.getBlockIndex();
	LineRecord[] list = linesRecord.getList();
	if (blockIndex >= blocks.size()) {
	    throw new CorruptedFileException();
	}
	LineEntryList entryList = new LineEntryList(sourceFile);
	for (LineRecord rec : list) {
	    int number = rec.getNumber();
	    if (number == 0) {
		entryList.changeFileName(rec.getFileName());
	    } else {
		entryList.addLineNumber(number);
	    }
	}
	blocks.get(blockIndex).setLines(entryList.getLineEntries());
    }

    /**
       եդ򤭤ޤ

       @throws CorruptedFileException եι¤Ƥ뤳Ȥ򸡽
    */
    private void solveFlowGraph() throws CorruptedFileException {
	Solver s = new Solver();
	for (T e : blocks) {
	    e.sortOutArcs();
	    s.addInvalid(e);
	}
	s.solve();
    }

    /**
       ؿθƤӽФ¹Ԥ줿֥åȽи
       ˤ׻ޤ

       calledCount, returnedCount, executedBlockCountͭˤʤޤ
    */
    private void countCallSummary() {
	calledCount = blocks.get(0).getCount();

	int blockCount = blocks.size();
	ArrayList<? extends AbstractArc> list
	    = blocks.get(blockCount - 1).getInArcs();
	long count = 0;
	for (AbstractArc arc : list) {
	    if (arc.isFake()) {
		continue;
	    }
	    count += arc.getCount();
	}
	returnedCount = count;

	int start = 1;
	int end = blockCount - 1;
	for (int k = start; k < end; ++k) {
	    if (blocks.get(k).getCount() > 0) {
		++executedBlockCount;
	    }
	}
    }

    /**
       ؿǡ쥳ɤؿդɲäޤ

       @param rec ؿǡ쥳
       @throws CorruptedFileException եι¤Ƥ뤳Ȥ򸡽
    */
    public final void setFunctionDataRecord(final FunctionDataRecord rec)
	throws CorruptedFileException {
	if (checksum != rec.getChecksum()) {
	    String m = String.format("gcda file: checksum mismatch for '%s'",
				     functionName);
	    throw new CorruptedFileException(m);
	}
	long[] arcCounts = rec.getArcCounts();
	if (solvedArcs.size() != arcCounts.length) {
	    String m = String.format("gcda file: profile mismatch for '%s'",
				     functionName);
	    throw new CorruptedFileException(m);
	}
	for (int k = 0; k < arcCounts.length; ++k) {
	    solvedArcs.get(k).addCount(arcCounts[k]);
	}
	solveFlowGraph();
	countCallSummary();
	solved = true;
    }

    /**
       ̻Ҥޤ

       @return ̻
    */
    public final int getId() {
	return id;
    }

    /**
       ؿϤޤֹޤ

       @return ؿϤޤֹ
    */
    public final int getLineNumber() {
	return lineNumber;
    }

    /**
       ؿ̾ޤ

       @return ؿ̾
    */
    public final String getFunctionName() {
	return functionName;
    }

    /**
       ؿƤФ줿ޤ

       setFunctionDataRecord()Ǵؿǡ쥳ɤꤵƤ
       ɬפޤ

       @return ؿƤФ줿
    */
    public final long getCalledCount() {
	return calledCount;
    }

    /**
       ؿäޤ

       setFunctionDataRecord()Ǵؿǡ쥳ɤꤵƤ
       ɬפޤ

       @return ؿä
    */
    public final long getReturnedCount() {
	return returnedCount;
    }

    /**
       и¹Ԥ줿֥åޤ

       setFunctionDataRecord()Ǵؿǡ쥳ɤꤵƤ
       ɬפޤ

       @return ¹Ԥ줿֥åи
    */
    public final int getExecutedBlockCount() {
	return executedBlockCount;
    }

    /**
       и֥åޤ

       @return ֥åи
    */
    public final int getBlockCount() {
	return blocks.size() - 2;
    }
}
